#include <iostream>
#include <fstream>
#include <cmath>
#include <cstring>
#include <vector>

#include "Std.h"
#include "Pset.h"
#include "Astrometry.h"
#include "Image.h"
#include "StringStuff.h"
#include "HeaderFromStream.h"
#include "SCAMPMap.h"
#include "FITSImage.h"

using namespace std;
using namespace astrometry;
using namespace stringstuff;


struct outlier {
    int id; // 0 ... n-1
    int x;
    int y;
    float sigma;
};

struct mask { // rectangular mask
    int x0;
    int x1;
    int y0;
    int y1;
};

struct filter { // rectangular filter
    int boxwidth;
    int threshcount;
    double threshold;
    static bool compareBySize(const filter &a, const filter &b)
    {
	return (a.boxwidth > b.boxwidth);
    }
};

int main(int argc, char **argv)
{
  string fstackhead, fmask, foutliers, fheadlist, masksuffix;
  int xsize,ysize;
  
  Pset parameters;
  {
      const int def=PsetMember::hasDefault;
      const int low=PsetMember::hasLowerBound;
      const int up=PsetMember::hasUpperBound;
      const int lowopen = low | PsetMember::openLowerBound;
      const int upopen = up | PsetMember::openUpperBound;

      parameters.addMember("outliers",&foutliers, def,
			   "outlier list as producted by SWarp", "clipped.tab");
      parameters.addMember("stackhead",&fstackhead, def,
			   "WCS header of stack, ASCII format", "stack.head");
      parameters.addMember("headlist",&fheadlist, def,
			   "list of single frame headers (in same order as input to SWarp)", "headlist.tab");
      parameters.addMember("mask",&fmask, def,
			   "file with mask definitions in format THRESHOLD_COUNT THRESHOLD_LEVEL[sigma] BOX_WIDTH", "mask.conf");
      parameters.addMember("masksuffix",&masksuffix, def,
			   "suffix for mask files", ".mask.fits");
      parameters.addMember("xsize", &xsize, def | low, "x size of single frames", 2048, 1); 
      parameters.addMember("ysize", &ysize, def | low, "y size of single frames", 4096, 1); 

  }

  parameters.setDefault();

  cout << "# " << argv[0] << " is reading parameters from config files passed to command line and stdin (Ctrl+D to finish)" << endl;
  
  for (int i=1; i<argc; i++) {
    // Open & read all specified input files
    ifstream ifs(argv[i]);
    if (!ifs) {
      cerr << "Can't open file " << argv[i] << endl;
      exit(1);
    }
    try {
      parameters.setStream(ifs);
    } catch (std::runtime_error &m) {
      cerr << "In file " << argv[i] << ":" << endl;
      quit(m,1);
    }
  }
  // and stdin:
  try {
    parameters.setStream(cin);
  } catch (std::runtime_error &m) {
    cerr << "In stdin:" << endl;
    quit(m,1);
  }


  cout << "# " << stringstuff::taggedCommandLine(argc, argv) << endl;
  parameters.dump(cout);

  // hierarchical masking
  vector<filter> filters;
  ifstream maskconf(fmask.c_str());
  filter f;
  while(maskconf >> f.boxwidth >> f.threshold >> f.threshcount)
    {
      filters.push_back(f);
    }
  sort(filters.begin(), filters.end(), filter::compareBySize);
  int nmasks = filters.size();
  
  cerr << "# read filters: ";
  for (int i=0; i<nmasks; i++)
  {
    cerr << filters[i].threshcount << "pix " << filters[i].threshold << "sig off in " << filters[i].boxwidth << "^2 box; ";
  }
  cerr << endl;
  
  // (-1) read stack WCS header
  SCAMPMap *stackmap=0;
  {
  ifstream mapfs(fstackhead.c_str());
  if (!mapfs) {
    cerr << "Could not open map spec file " << fstackhead << endl;
    exit(1);
  }

  // Set up SCAMPMap to go to reprojected center if desired
  img::ImageHeader h = img::HeaderFromStream(mapfs);
  stackmap = new SCAMPMap(h);
  if (!stackmap) {
    cerr << "Error reading input WCS map from file " << fstackhead
	     << endl;
    exit(1);
  }
  }
  
  cerr << "# read stack header" << endl;
  
  // (0) read outlier list into vector of outliers
  int maxid=0;
  vector<outlier> outlierlist;
  double minthreshsigma = 1.e99;
  for (int i=0; i<nmasks; i++)
  {
      if(minthreshsigma>filters[i].threshold) minthreshsigma=filters[i].threshold;
  }
  
  {
    ifstream inoutlierlist(foutliers.c_str());
    float id,x,y,sigma;
    while(inoutlierlist >> id >> x >> y >> sigma)
    {	
	outlier tmp;
        tmp.id=int(id);
        if(tmp.id>maxid) maxid=tmp.id;
	tmp.x=int(x)+1;
	tmp.y=int(y)+1;
	tmp.sigma=sigma;
	if(sigma>minthreshsigma || sigma<-minthreshsigma) {
		outlierlist.push_back(tmp);
        }
    }	
  }
   
  vector<int> minx (maxid+1,INT_MAX);
  vector<int> maxx (maxid+1,-INT_MAX);
  vector<int> miny (maxid+1,INT_MAX);
  vector<int> maxy (maxid+1,-INT_MAX);
  vector<int> noutliers (maxid+1,0);

  for(int j=0; j<outlierlist.size(); j++)
  {
	noutliers[outlierlist[j].id]++;
        if(minx[outlierlist[j].id]>outlierlist[j].x) minx[outlierlist[j].id]=outlierlist[j].x;
        if(miny[outlierlist[j].id]>outlierlist[j].y) miny[outlierlist[j].id]=outlierlist[j].y;
        if(maxx[outlierlist[j].id]<outlierlist[j].x) maxx[outlierlist[j].id]=outlierlist[j].x;
        if(maxy[outlierlist[j].id]<outlierlist[j].y) maxy[outlierlist[j].id]=outlierlist[j].y;
  }

  // (1) read single frame headers
  cerr << "# read outlier list\n# reading single frame headers" << flush;
  vector<SCAMPMap *> map;
  vector<string> outmask;
  
  {
  ifstream headerlist(fheadlist.c_str());
  string hfile;
  int i=0;
  while(i<=maxid && (headerlist >> hfile))
  {
    ifstream mapfs(hfile.c_str());
    if (!mapfs) {
      cerr << "Could not open map spec file " << hfile << endl;
      exit(1);
    }

    // Set up SCAMPMap to go to reprojected center
    img::ImageHeader h = img::HeaderFromStream(mapfs);
    SCAMPMap *m = new SCAMPMap(h, &(stackmap->projection()));
    if (!stackmap) {
      cerr << "Error reading input WCS map from file " << hfile << endl;
      exit(1);
    }
    map.push_back(m); i++;
    outmask.push_back(hfile+masksuffix);
    cerr << "." << flush;
  }
  }
  
  cerr << "\n# read single frame headers" << endl;

  for(int i=0; i<=maxid; i++) // iterate all frames
  {
  
    cerr << "# " << "frame " << i << endl;
    if(!noutliers[i]) {
      cerr << "# no outliers in this frame" << endl;
      continue;    
    }
    int xdim = maxx[i]-minx[i];
    int ydim = maxy[i]-miny[i];

    vector<mask> allmasks;
    
    for(int n=0; n<nmasks; n++) // iterate all mask sizes
    {
	int pmax = (xdim+filters[n].boxwidth)*(ydim+filters[n].boxwidth);
	unsigned int *outliercounta = new unsigned int[pmax]; // outlier count (above) in boxes with upper-right corner; initialize to zero below
	unsigned int *outliercountb = new unsigned int[pmax]; // outlier count (below) in boxes with upper-right corner; initialize to zero below

	memset(outliercounta, 0, pmax*sizeof(unsigned int));
	memset(outliercountb, 0, pmax*sizeof(unsigned int));

	cout << "# " << filters[n].boxwidth << "x" << filters[n].boxwidth << " filter\n# (1) count outliers in boxes" << endl;

	for(int j=0; j<outlierlist.size(); j++)
	{
		if(outlierlist[j].id==i)
		{
			if(outlierlist[j].sigma>-filters[n].threshold && outlierlist[j].sigma<filters[n].threshold) // not above threshold
			  continue;
			// find correct sign buffer
			unsigned int *outliercount = (outlierlist[j].sigma>0)?outliercounta:outliercountb;

			// transform to the shifted coordinate system of the counting map
			int x0 = outlierlist[j].x-minx[i];
			int y0 = outlierlist[j].y-miny[i];

			for(int yt=y0; yt<y0+filters[n].boxwidth; yt++)
			  for(int xt=x0; xt<x0+filters[n].boxwidth; xt++) {
			    outliercount[xt+yt*(xdim+filters[n].boxwidth)]++;
                          }
		}
	}

	cout << "# (2) mask boxes" << endl;

	vector<mask> masklist;
	for(int j=0; j<pmax; j++) {
		if(outliercounta[j]>=filters[n].threshcount || outliercountb[j]>=filters[n].threshcount) { 
		// this is the upper-right corner of a masked square in the omax size frame
			int j0 = j;
			int j1 = j0;
			int x0 = j%(xdim+filters[n].boxwidth);
			int x1 = x0;
			int y0 = (j-x0)/(xdim+filters[n].boxwidth);
			int y1 = y0;
			int mask_offset=0; // how many pixel have we shifted since the last square legally masked
			do {
				j1++;
				x1 = j1%(xdim+filters[n].boxwidth);
				y1 = (j1-x1)/(xdim+filters[n].boxwidth);
				if(y1!=y0)
				{
					j1--;
					break;
				}
				if(outliercounta[j1]<filters[n].threshcount && outliercountb[j1]<filters[n].threshcount)
					mask_offset++;
				else
					mask_offset=0;
			} while(y1==y0 && mask_offset<filters[n].boxwidth);
			// we always go one too far...
			j1-=mask_offset;
			x1=j1%(xdim+filters[n].boxwidth);
			y1 = (j1-x1)/(xdim+filters[n].boxwidth);
			
			mask m;
			m.x0=x0-filters[n].boxwidth+minx[i]; m.x1=x1+minx[i]; m.y0=y0-filters[n].boxwidth+miny[i]; m.y1=y1+miny[i];
			masklist.push_back(m);
			j=j1;
		}
	}

	cout << "# (3) merge masks" << endl;

	// delete masks which are already covered
	if(allmasks.size()>0)
	{
	  for(int k=0; k<masklist.size(); k++)
	  {
		int x0=masklist[k].x0;
		int x1=masklist[k].x1;
		int y0=masklist[k].y0;
		int y1=masklist[k].y1;

		for(int l=0; l<allmasks.size(); l++)
		{
		 if(x0>=allmasks[l].x0 && x1<=allmasks[l].x1 && y0>=allmasks[l].x0 && y1<=allmasks[l].x1) // already covered!
		 {
		  masklist.erase(masklist.begin() + k);
		  k--;
		  goto nextplease;
		 }
		 if(allmasks[l].y0>y0)
		  goto nextplease;
		}
		
		nextplease:
		;
	  }
	}

	// merge masks; use the fact they are ordered by y0
	for(int k=0; k<masklist.size(); k++)
	{
		int x0=masklist[k].x0;
		int x1=masklist[k].x1;
		int y0=masklist[k].y0;
		int y1=masklist[k].y1;

		for(int j=k+1; j<masklist.size(); j++)
		{
			if(masklist[j].y0>y1) // we're out
				break;

			if(x0==masklist[j].x0 && x1==masklist[j].x1) // aligned mask
			{
				masklist[k].y1=y1=masklist[j].y1; // expand first mask
				masklist.erase(masklist.begin() + j); // erase merged mask
				j--; // we do not want to forget about the new j'th mask
			}
		}
	}
	
	// fill masks into allmask-list
	allmasks.insert(allmasks.end(), masklist.begin(), masklist.end());
	
	delete []outliercounta;
	delete []outliercountb;
	
    }
	
    cerr << "# (4) generating mask image" << endl;
	
	
    Image<float> mask(Bounds<int>(1,xsize,1,ysize),float(1));
    
    int npixmasked=0;
    for(int k=0; k<allmasks.size(); k++)
    {
	    double xw,yw,xo,yo,xomin,xomax,yomin,yomax;
	    stackmap->toWorld(allmasks[k].x0, allmasks[k].y0, xw, yw);
	    map[i]->toPix(xw, yw, xo, yo);
	    xomin=xomax=xo; yomin=yomax=yo;
	   
	    stackmap->toWorld(allmasks[k].x0, allmasks[k].y1, xw, yw);
	    map[i]->toPix(xw, yw, xo, yo);
	    if(xo<xomin) xomin=xo;
	    if(xo>xomax) xomax=xo;
	    if(yo<yomin) yomin=yo;
	    if(yo>yomax) yomax=yo;
	    
	    stackmap->toWorld(allmasks[k].x1, allmasks[k].y1, xw, yw);
	    map[i]->toPix(xw, yw, xo, yo);
	    if(xo<xomin) xomin=xo;
	    if(xo>xomax) xomax=xo;
	    if(yo<yomin) yomin=yo;
	    if(yo>yomax) yomax=yo;
	    
	    stackmap->toWorld(allmasks[k].x1, allmasks[k].y0, xw, yw);
	    map[i]->toPix(xw, yw, xo, yo);
	    if(xo<xomin) xomin=xo;
	    if(xo>xomax) xomax=xo;
	    if(yo<yomin) yomin=yo;
	    if(yo>yomax) yomax=yo;
	    
	    if(xomin<0) xomin=0;
	    if(yomin<0) yomin=0;
	    if(xomax>xsize-2) xomax=xsize-2;
	    if(yomax>ysize-2) yomax=ysize-2;
	    
	    for(int ix=int(xomin)+1; ix<=int(xomax)+2; ix++)
	    for(int iy=int(yomin)+1; iy<=int(yomax)+2; iy++)
	    {
	      if(mask(ix,iy)>0) {
		npixmasked++;
	        mask(ix,iy)=0.;
	      }
	    }
    }

    FITSImage<>::writeToFITS(string(outmask[i]),mask);
    cout << outmask[i] << " " << double(npixmasked)/double(xsize*ysize) << endl;

  }
  
  return 0;
}
