#include <iostream>
#include <cstring>

#include "SCAMPMap.h"
#include "HeaderFromStream.h"
#include "SBPixel.h"
#include "PSFEx.h"
#include "Pset.h"

using namespace std;
using namespace astrometry;
using namespace sbp;

const int Nsquare=30;  // size of postage stamp

double GetMedian(double daArray[], int Nframes) {
    // Allocate an array of the same size and sort it.
    double* dpSorted = new double[Nframes];
    for (int i = 0; i < Nframes; ++i) {
        dpSorted[i] = daArray[i];
    }
    for (int i = Nframes - 1; i > 0; --i) {
        for (int j = 0; j < i; ++j) {
            if (dpSorted[j] > dpSorted[j+1]) {
                double dTemp = dpSorted[j];
                dpSorted[j] = dpSorted[j+1];
                dpSorted[j+1] = dTemp;
            }
        }
    }

    // Middle or average of middle values in the sorted array.
    double dMedian = 0.0;
    if ((Nframes % 2) == 0) {
        dMedian = (dpSorted[Nframes/2] + dpSorted[(Nframes/2) - 1])/2.0;
    } else {
        dMedian = dpSorted[Nframes/2];
    }
    delete [] dpSorted;
    return dMedian;
}

double max(double a, int b)
{
 if(a>b) return a;
 return b; 
}

int main(int argc, char **argv)
{
   
  double flux,gain,x,y;
  string fstackhead, fheadlist, fpsflist; 
  
  Pset parameters;
  {
      const int def=PsetMember::hasDefault;
      const int low=PsetMember::hasLowerBound;
      const int up=PsetMember::hasUpperBound;
      const int lowopen = low | PsetMember::openLowerBound;
      const int upopen = up | PsetMember::openUpperBound;

      
      parameters.addMember("flux", &flux, def | low, "total flux of star in units of background noise rms", 1000., 1.); 
      parameters.addMember("gain", &gain, def | low, "typical product of gain and background noise rms for single frames; 0 to assume infinite gain (no object-dependent shot noise)", 0., 0.); 
      parameters.addMember("stackhead",&fstackhead, def,
			   "WCS header of stack, ASCII format", "stack.head");
      parameters.addMember("headlist",&fheadlist, def,
			   "list of single frame headers (in same order as input to SWarp)", "headlist.tab");
      parameters.addMember("psflist",&fpsflist, def,
			   "list of single frame PSFEx models (in same order as input to SWarp)", "psflist.tab");
      parameters.addMember("x", &x, def | low, "x position in stack at which PSF homogeneity is tested", 1000., 1.); 
      parameters.addMember("y", &y, def | low, "y position in stack at which PSF homogeneity is tested", 1000., 1.); 

  }
  
  parameters.setDefault();

  cout << "# " << argv[0] << " is reading parameters from config files passed to command line and stdin (Ctrl+D to finish)" << endl;
  
  for (int i=1; i<argc; i++) {
    // Open & read all specified input files
    ifstream ifs(argv[i]);
    if (!ifs) {
      cerr << "Can't open file " << argv[i] << endl;
      exit(1);
    }
    try {
      parameters.setStream(ifs);
    } catch (std::runtime_error &m) {
      cerr << "In file " << argv[i] << ":" << endl;
      quit(m,1);
    }
  }
  // and stdin:
  try {
    parameters.setStream(cin);
  } catch (std::runtime_error &m) {
    cerr << "In stdin:" << endl;
    quit(m,1);
  }


  cout << "# " << stringstuff::taggedCommandLine(argc, argv) << endl;
  parameters.dump(cout);
    
  // (-1) read stack WCS header
  SCAMPMap *stackmap=0;
  {
  ifstream mapfs(fstackhead.c_str());
  if (!mapfs) {
    cerr << "Could not open map spec file " << fstackhead << endl;
    exit(1);
  }

  // Set up SCAMPMap to go to reprojected center if desired
  img::ImageHeader h = img::HeaderFromStream(mapfs);
  stackmap = new SCAMPMap(h);
  if (!stackmap) {
    cerr << "Error reading input WCS map from file " << fstackhead
	     << endl;
    exit(1);
  }
  }
  
  cerr << "# finished reading stack header\nreading single frame headers" << flush;
  
  vector<SCAMPMap *> map;

  {
    ifstream headerlist(fheadlist.c_str());
    string hfile;

    while(headerlist >> hfile)
    {
      ifstream mapfs(hfile.c_str());
      if (!mapfs) {
	cerr << "Could not open map spec file " << hfile << endl;
	exit(1);
      }

      // Set up SCAMPMap to go to reprojected center
      img::ImageHeader h = img::HeaderFromStream(mapfs);
      SCAMPMap *m = new SCAMPMap(h, &(stackmap->projection()));
      if (!m) {
	cerr << "Error reading input WCS map from file " << hfile << endl;
	exit(1);
      }
      
      map.push_back(m);
      cerr << "." << flush;
    }
  }
  
  int Nframes = map.size();
  
  if(Nframes<2) {
   cerr << "cannot do with fewer than 2 single frames; exiting." << endl;
   exit(1);
  }
  
  cerr << "\n# finished reading " << Nframes << " single frame headers\nreading single frame PSF models" << flush;
  
  vector<PSFExModel *> psf; 
  {
  ifstream psfmodellist(fpsflist.c_str());
  string pfile;

  while(psfmodellist >> pfile)
  {
    PSFExModel *p = new PSFExModel(pfile.c_str());
    if(!p) {
       cerr << "Error reading PSF model from file " << pfile << endl;
       exit(1);
    }
    
    psf.push_back(p);
    cerr << "." << flush;
  }
  }
  
  if(psf.size() != Nframes)
  {
    cerr << "Number of single frame headers and PSF models do not agree. Exiting." << endl;
    exit(1);
  }
  
  cerr << "\n# finished reading single frame PSF models" << endl;
    
  cerr << "# drawing PSF images" << endl;
  
  double xw,yw,xo,yo;
  stackmap->toWorld(x, y, xw, yw);
  
  for(int i=0; i<Nframes; i++) // set position in image
  {
	    map[i]->toPix(xw, yw, xo, yo);
	   
	    if(!psf[i]->isInsideInterpolatedRegion(xo,yo)) { // PSF is extrapolated outside frame, do not use
	      cerr << "ignoring one single frame PSF model which is invalid at given stack position" << endl;
	      map.erase(map.begin()+i);
	      psf.erase(psf.begin()+i);
	      Nframes--;
	      i--;
	      continue;
	    }
	    psf[i]->fieldPosition(xo,yo);
  }
  
  if(Nframes<2) {
    cerr << "cannot do with fewer than 2 single frame PSFs; exiting." << endl;
    exit(1);  
  }
  
  cerr << "# " << Nframes << " are overlapping at the specified position" << endl;
  
  double image[Nsquare][Nsquare][Nframes];
  double medim[Nsquare][Nsquare];
  double outlier[Nsquare][Nsquare][Nframes];

  for(int x=0; x<Nsquare; x++)
  {
    for(int y=0; y<Nsquare; y++)
    {
     Position<double> pos(double(x-Nsquare/2.),double(y-Nsquare/2.));
       
     for(int i=0; i<Nframes; i++)
     {
       image[x][y][i] = psf[i]->sb()->xValue(pos);
     }
    }
  }
   
  cerr << "// get median of each pixel" << endl;
  for(int x=0; x<Nsquare; x++) 
  for(int y=0; y<Nsquare; y++)
    {
	medim[x][y]=GetMedian(image[x][y], Nframes);
    }  
  
  cerr << "// generate statistics" << endl;
  for(double A=0.; A<0.8; A+=0.05)
  {
    // generate outlier maps
    for(int x=0; x<Nsquare; x++)
    {
    for(int y=0; y<Nsquare; y++)
    {
    double m = medim[x][y];
    for(int i=0; i<Nframes; i++)
    {
        double im = image[x][y][i];
	double d = im-m;
	outlier[x][y][i] = ((d>0)?(1.):(-1.)) * fabs(max(fabs(d)-A*fabs(m),0));
    }
    }
    }
    
    int Nmax[Nframes][2];
    
    // get Nmax as a function of n
    for(double n=2.5; n<10; n+=0.01)
    {
     memset(Nmax, 0, sizeof(int)*Nframes*2);
     
     // generate outlier maps
     if(gain<=0.) {
      for(int x=0; x<Nsquare; x++)
      {
      for(int y=0; y<Nsquare; y++)
      {
      for(int i=0; i<Nframes; i++)
      {
	  if(outlier[x][y][i]>n/flux) Nmax[i][0]++;
	  else if(outlier[x][y][i]<-n/flux) Nmax[i][1]++;
      }
      }
      }
     }
     else {
      for(int x=0; x<Nsquare; x++)
      {
      for(int y=0; y<Nsquare; y++)
      {
      double shotnoiseboost = sqrt(1.+medim[x][y]*flux/gain);
      for(int i=0; i<Nframes; i++)
      {
	  if(outlier[x][y][i]>n/flux*shotnoiseboost) Nmax[i][0]++;
	  else if(outlier[x][y][i]<-n/flux*shotnoiseboost) Nmax[i][1]++;
      }
      }
      } 
       
     }
     
     int N=0;
     for(int i=0; i<Nframes; i++)
       for(int j=0; j<2; j++)
       {
	if(Nmax[i][j]>N) N=Nmax[i][j]; 
       }
       
     cout << A << " " << n << " " << N << endl;
    }
  }
    
  return 0;
}
